"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.visitSourceFiles = void 0;
var model_1 = require("./model");
var ts_morph_1 = require("ts-morph");
var metric_collector_1 = require("./metric_collector");
// We push 'null' on the stack to signal children of local variables not relevant for Sonargraph
var parentStack = [];
var localSymbols = new Set();
var emptySet = new Set();
var root;
var delayedActions = [];
function getCurrentParent() {
    return parentStack[parentStack.length - 1];
}
function getLastNonNullParent() {
    for (var i = parentStack.length - 1; i >= 0; i--) {
        var parent_1 = parentStack[i];
        if (parent_1)
            return parent_1;
    }
    throw new Error("Parent stack only contains null pointers");
}
function handleParameterTypes(fd, parameters, recordTypeDependencies, returnType) {
    if (recordTypeDependencies === void 0) { recordTypeDependencies = true; }
    parameters.forEach(function (param) {
        var sym = param.getSymbol();
        if (sym) {
            localSymbols.add(sym);
        }
        if (recordTypeDependencies)
            addDependencyToType(fd, param.getType(), new Set(), param.getStartLineNumber());
    });
    if (returnType) {
        addDependencyToType(fd, returnType.getType(), new Set(), returnType.getStartLineNumber());
    }
}
function handlePropertyOrVariable(v, typeNode) {
    addDependencyToType(v, typeNode.getType(), new Set(), typeNode.getStartLineNumber());
}
function handleParametersAndReturnType(fd, params, typeParams, returnTypeNode) {
    var paramsAsString = [];
    var typeParamsAsString = [];
    var returnType = "void";
    params.forEach(function (pd) { return paramsAsString.push(pd.getText()); });
    typeParams.forEach(function (tp) { return typeParamsAsString.push(tp.getText()); });
    if (returnTypeNode)
        returnType = returnTypeNode.getText();
    fd.numberOfParameters = params.length;
    fd.parameters = paramsAsString.join(";");
    fd.typeParameters = typeParamsAsString.join(";");
    fd.returnType = returnType;
}
var FunctionDeclarationHandler = /** @class */ (function () {
    function FunctionDeclarationHandler() {
    }
    FunctionDeclarationHandler.prototype.enter = function (node) {
        var sym = node.getSymbol();
        var fd;
        var parent = getCurrentParent();
        if (!parent || !sym)
            fd = null;
        else if (node.getBody())
            fd = new model_1.FunctionOrMethodWithBody(parent, sym.getName(), node, node.getStartLineNumber(), model_1.ElementType.Function, node.getBody());
        else
            fd = new model_1.FunctionOrMethodWithoutBody(parent, sym.getName(), node, node.getStartLineNumber(), model_1.ElementType.Function, false);
        if (fd) {
            handleParametersAndReturnType(fd, node.getParameters(), node.getTypeParameters(), node.getReturnTypeNode());
            delayedActions.push(function () { return handleParameterTypes(fd, node.getParameters(), true, node.getReturnTypeNode()); });
        }
        parentStack.push(fd);
    };
    return FunctionDeclarationHandler;
}());
var ConstructorDeclarationHandler = /** @class */ (function () {
    function ConstructorDeclarationHandler() {
    }
    ConstructorDeclarationHandler.prototype.enter = function (node) {
        var sym = node.getSymbol();
        var parent = getCurrentParent();
        if (parent && sym) {
            var fd_1;
            if (node.getBody())
                fd_1 = new model_1.FunctionOrMethodWithBody(parent, sym.getName(), node, node.getStartLineNumber(), model_1.ElementType.Constructor, node.getBody());
            else
                fd_1 = new model_1.FunctionOrMethodWithoutBody(parent, sym.getName(), node, node.getStartLineNumber(), model_1.ElementType.Constructor, false);
            delayedActions.push(function () { return handleParameterTypes(fd_1, node.getParameters()); });
            handleParametersAndReturnType(fd_1, node.getParameters(), node.getTypeParameters(), undefined);
            parentStack.push(fd_1);
        }
        else
            parentStack.push(null);
    };
    return ConstructorDeclarationHandler;
}());
var MethodDeclarationHandler = /** @class */ (function () {
    function MethodDeclarationHandler() {
    }
    MethodDeclarationHandler.prototype.enter = function (node) {
        var sym = node.getSymbol();
        var fd;
        var parent = getCurrentParent();
        if (parent && sym) {
            if (node.isAbstract() || !node.getBody())
                fd = new model_1.FunctionOrMethodWithoutBody(parent, sym.getName(), node, node.getStartLineNumber(), model_1.ElementType.Method, node.isAbstract());
            else
                fd = new model_1.FunctionOrMethodWithBody(parent, sym.getName(), node, node.getStartLineNumber(), model_1.ElementType.Method, node.getBody());
            handleParametersAndReturnType(fd, node.getParameters(), node.getTypeParameters(), node.getReturnTypeNode());
            delayedActions.push(function () { return handleParameterTypes(fd, node.getParameters(), true, node.getReturnTypeNode()); });
            parentStack.push(fd);
        }
        else
            parentStack.push(null);
    };
    return MethodDeclarationHandler;
}());
var MethodSignatureHandler = /** @class */ (function () {
    function MethodSignatureHandler() {
    }
    MethodSignatureHandler.prototype.enter = function (node) {
        var sym = node.getSymbol();
        var parent = getCurrentParent();
        if (parent && sym) {
            var fd_2 = new model_1.FunctionOrMethodWithoutBody(parent, sym.getName(), node, node.getStartLineNumber(), model_1.ElementType.Method, false);
            handleParametersAndReturnType(fd_2, node.getParameters(), node.getTypeParameters(), node.getReturnTypeNode());
            delayedActions.push(function () { return handleParameterTypes(fd_2, node.getParameters(), true, node.getReturnTypeNode()); });
            parentStack.push(fd_2);
        }
        else
            parentStack.push(null);
    };
    return MethodSignatureHandler;
}());
function handleInheritance(from, clauses) {
    for (var _i = 0, clauses_1 = clauses; _i < clauses_1.length; _i++) {
        var clause = clauses_1[_i];
        var isExtends = clause.getToken() == ts_morph_1.SyntaxKind.ExtendsKeyword;
        for (var _a = 0, _b = clause.getTypeNodes(); _a < _b.length; _a++) {
            var tn = _b[_a];
            addDependencyToType(from, tn.getType(), new Set(), tn.getStartLineNumber(), isExtends ? model_1.DependencyKind.EXTENDS : model_1.DependencyKind.IMPLEMENTS);
        }
    }
}
var ClassDeclarationHandler = /** @class */ (function () {
    function ClassDeclarationHandler() {
    }
    ClassDeclarationHandler.prototype.enter = function (node) {
        var sym = node.getSymbol();
        var parent = getCurrentParent();
        if (parent && sym) {
            var cls_1 = new model_1.Class(parent, sym.getName(), node, node.getStartLineNumber(), node.isAbstract());
            var typeParamsAsString = [];
            for (var _i = 0, _a = node.getTypeParameters(); _i < _a.length; _i++) {
                var tp = _a[_i];
                var sym_1 = tp.getSymbol();
                if (sym_1)
                    localSymbols.add(sym_1);
                typeParamsAsString.push(tp.getText());
            }
            cls_1.typeParameters = typeParamsAsString.join(";");
            delayedActions.push(function () { return handleInheritance(cls_1, node.getHeritageClauses()); });
            parentStack.push(cls_1);
        }
        else
            parentStack.push(null);
    };
    return ClassDeclarationHandler;
}());
var InterfaceDeclarationHandler = /** @class */ (function () {
    function InterfaceDeclarationHandler() {
    }
    InterfaceDeclarationHandler.prototype.enter = function (node) {
        var sym = node.getSymbol();
        var parent = getCurrentParent();
        if (parent && sym) {
            var i_1 = new model_1.Interface(parent, sym.getName(), node, node.getStartLineNumber());
            var typeParamsAsString = [];
            for (var _i = 0, _a = node.getTypeParameters(); _i < _a.length; _i++) {
                var tp = _a[_i];
                var sym_2 = tp.getSymbol();
                if (sym_2)
                    localSymbols.add(sym_2);
                typeParamsAsString.push(tp.getText());
            }
            i_1.typeParameters = typeParamsAsString.join(";");
            delayedActions.push(function () { return handleInheritance(i_1, node.getHeritageClauses()); });
            parentStack.push(i_1);
        }
        else
            parentStack.push(null);
    };
    return InterfaceDeclarationHandler;
}());
var EnumDeclarationHandler = /** @class */ (function () {
    function EnumDeclarationHandler() {
    }
    EnumDeclarationHandler.prototype.enter = function (node) {
        var sym = node.getSymbol();
        var parent = getCurrentParent();
        if (parent && sym) {
            var e = new model_1.Enum(parent, sym.getName(), node, node.getStartLineNumber());
            parentStack.push(e);
        }
        else
            parentStack.push(null);
    };
    return EnumDeclarationHandler;
}());
var PropertyDeclarationHandler = /** @class */ (function () {
    function PropertyDeclarationHandler() {
    }
    PropertyDeclarationHandler.prototype.enter = function (node) {
        var sym = node.getSymbol();
        if (sym) {
            var parent_2 = getCurrentParent();
            if (parent_2) {
                var prop_1 = new model_1.PropertyOrVariable(parent_2, sym.getName(), node, node.getStartLineNumber(), model_1.ElementType.Property, node.getInitializer());
                parentStack.push(prop_1);
                if (node.getTypeNode())
                    delayedActions.push(function () { return handlePropertyOrVariable(prop_1, node.getTypeNode()); });
            }
            else
                parentStack.push(null);
        }
        else
            parentStack.push(null);
    };
    return PropertyDeclarationHandler;
}());
var PropertySignatureHandler = /** @class */ (function () {
    function PropertySignatureHandler() {
    }
    PropertySignatureHandler.prototype.enter = function (node) {
        var sym = node.getSymbol();
        if (sym) {
            var parent_3 = getCurrentParent();
            if (parent_3) {
                var prop_2 = new model_1.PropertyOrVariable(parent_3, sym.getName(), node, node.getStartLineNumber(), model_1.ElementType.Property, node.getInitializer());
                parentStack.push(prop_2);
                if (node.getTypeNode())
                    delayedActions.push(function () { return handlePropertyOrVariable(prop_2, node.getTypeNode()); });
            }
            else
                parentStack.push(null);
        }
        else
            parentStack.push(null);
    };
    return PropertySignatureHandler;
}());
var VariableDeclarationHandler = /** @class */ (function () {
    function VariableDeclarationHandler() {
    }
    VariableDeclarationHandler.prototype.enter = function (node) {
        var sym = node.getSymbol();
        var parent = getCurrentParent();
        var from;
        if (sym) {
            if (parent == null || parent instanceof model_1.FunctionOrMethodWithBody) {
                // Don't add local variables to the model
                localSymbols.add(sym);
                from = getLastNonNullParent();
                parentStack.push(null);
            }
            else {
                var v = new model_1.PropertyOrVariable(parent, sym.getName(), node, node.getStartLineNumber(), model_1.ElementType.Variable, node.getInitializer());
                parentStack.push(v);
                from = v;
            }
            if (node.getTypeNode())
                delayedActions.push(function () { return handlePropertyOrVariable(from, node.getTypeNode()); });
        }
        else
            parentStack.push(null);
    };
    return VariableDeclarationHandler;
}());
var GetAccessorHandler = /** @class */ (function () {
    function GetAccessorHandler() {
    }
    GetAccessorHandler.prototype.enter = function (node) {
        var sym = node.getSymbol();
        var parent = getCurrentParent();
        if (parent && sym) {
            var getter_1;
            if (node.isAbstract() || !node.getBody())
                getter_1 = new model_1.FunctionOrMethodWithoutBody(parent, sym.getName(), node, node.getStartLineNumber(), model_1.ElementType.Getter, node.isAbstract());
            else
                getter_1 = new model_1.FunctionOrMethodWithBody(parent, sym.getName(), node, node.getStartLineNumber(), model_1.ElementType.Getter, node.getBody());
            handleParametersAndReturnType(getter_1, node.getParameters(), node.getTypeParameters(), node.getReturnTypeNode());
            delayedActions.push(function () { return handleParameterTypes(getter_1, node.getParameters(), true, node.getReturnTypeNode()); });
            parentStack.push(getter_1);
        }
        else
            parentStack.push(null);
    };
    return GetAccessorHandler;
}());
var SetAccessorHandler = /** @class */ (function () {
    function SetAccessorHandler() {
    }
    SetAccessorHandler.prototype.enter = function (node) {
        var sym = node.getSymbol();
        var parent = getCurrentParent();
        if (parent && sym) {
            var setter_1;
            if (node.getBody())
                setter_1 = new model_1.FunctionOrMethodWithBody(parent, sym.getName(), node, node.getStartLineNumber(), model_1.ElementType.Setter, node.getBody());
            else
                setter_1 = new model_1.FunctionOrMethodWithoutBody(parent, sym.getName(), node, node.getStartLineNumber(), model_1.ElementType.Setter, node.isAbstract());
            handleParametersAndReturnType(setter_1, node.getParameters(), node.getTypeParameters(), node.getReturnTypeNode());
            delayedActions.push(function () { return handleParameterTypes(setter_1, node.getParameters()); });
            parentStack.push(setter_1);
        }
        else
            parentStack.push(null);
    };
    return SetAccessorHandler;
}());
var EnumMemberHandler = /** @class */ (function () {
    function EnumMemberHandler() {
    }
    EnumMemberHandler.prototype.enter = function (node) {
        var sym = node.getSymbol();
        var parent = getCurrentParent();
        if (parent && sym) {
            var m = new model_1.PropertyOrVariable(parent, sym.getName(), node, node.getStartLineNumber(), model_1.ElementType.EnumMember, node.getInitializer());
            parentStack.push(m);
        }
        else
            parentStack.push(null);
    };
    return EnumMemberHandler;
}());
var NamespaceHandler = /** @class */ (function () {
    function NamespaceHandler() {
    }
    NamespaceHandler.prototype.enter = function (node) {
        var sym = node.getSymbol();
        var parent = getCurrentParent();
        if (parent && sym) {
            var ns = new model_1.Namespace(parent, sym.getName(), node, node.getStartLineNumber());
            parentStack.push(ns);
        }
        else
            parentStack.push(null);
    };
    return NamespaceHandler;
}());
var TypeAliasHandler = /** @class */ (function () {
    function TypeAliasHandler() {
    }
    TypeAliasHandler.prototype.enter = function (node) {
        var sym = node.getSymbol();
        var parent = getCurrentParent();
        if (parent && sym) {
            var ta_1 = new model_1.TypeAlias(parent, sym.getName(), node, node.getStartLineNumber());
            var typeParamsAsString_1 = [];
            node.getTypeParameters().forEach(function (tp) { return typeParamsAsString_1.push(tp.getText()); });
            ta_1.typeParameters = typeParamsAsString_1.join(";");
            delayedActions.push(function () { return addDependencyToType(ta_1, node.getType(), new Set(), node.getStartLineNumber()); });
            parentStack.push(ta_1);
        }
        else
            parentStack.push(null);
    };
    return TypeAliasHandler;
}());
var PropertyAssignmentHandler = /** @class */ (function () {
    function PropertyAssignmentHandler() {
    }
    PropertyAssignmentHandler.prototype.enter = function (node) {
        var parent = getCurrentParent();
        if (parent instanceof model_1.PropertyOrVariable) {
            var nameNode = node.getNameNode();
            if (isIdentifier(nameNode)) {
                var prop = new model_1.PropertyOrVariable(parent, nameNode.getText(), node, node.getStartLineNumber(), model_1.ElementType.Property, node.getInitializer());
                parentStack.push(prop);
            }
            else
                parentStack.push(null);
        }
        else
            parentStack.push(null);
    };
    return PropertyAssignmentHandler;
}());
var ClassExpressionHandler = /** @class */ (function () {
    function ClassExpressionHandler() {
    }
    ClassExpressionHandler.prototype.enter = function (node) {
        var parent = getCurrentParent();
        if (parent) {
            var cls_2 = new model_1.Class(parent, "", node, node.getStartLineNumber(), false);
            for (var _i = 0, _a = node.getTypeParameters(); _i < _a.length; _i++) {
                var tp = _a[_i];
                var sym = tp.getSymbol();
                if (sym)
                    localSymbols.add(sym);
            }
            delayedActions.push(function () { return handleInheritance(cls_2, node.getHeritageClauses()); });
            parentStack.push(cls_2);
        }
        else
            parentStack.push(null);
    };
    return ClassExpressionHandler;
}());
var FunctionExpressionHandler = /** @class */ (function () {
    function FunctionExpressionHandler() {
    }
    FunctionExpressionHandler.prototype.enter = function (node) {
        var fd;
        var parent = getCurrentParent();
        if (!parent)
            fd = null;
        else
            fd = new model_1.FunctionOrMethodWithBody(parent, "", node, node.getStartLineNumber(), model_1.ElementType.Function, node.getBody());
        if (fd) {
            handleParametersAndReturnType(fd, node.getParameters(), node.getTypeParameters(), node.getReturnTypeNode());
            delayedActions.push(function () { return handleParameterTypes(fd, node.getParameters(), true, node.getReturnTypeNode()); });
        }
        parentStack.push(fd);
    };
    return FunctionExpressionHandler;
}());
var ArrowFunctionHandler = /** @class */ (function () {
    function ArrowFunctionHandler() {
    }
    ArrowFunctionHandler.prototype.enter = function (node) {
        var parent = getCurrentParent();
        if (!parent)
            parent = getLastNonNullParent();
        var fd = new model_1.FunctionOrMethodWithBody(parent, "", node, node.getStartLineNumber(), model_1.ElementType.Function, node.getBody());
        handleParametersAndReturnType(fd, node.getParameters(), node.getTypeParameters(), node.getReturnTypeNode());
        delayedActions.push(function () { return handleParameterTypes(fd, node.getParameters(), true, node.getReturnTypeNode()); });
        parentStack.push(fd);
    };
    return ArrowFunctionHandler;
}());
function createHandler(node) {
    switch (node.getKind()) {
        case ts_morph_1.SyntaxKind.ClassExpression:
            return new ClassExpressionHandler();
        case ts_morph_1.SyntaxKind.FunctionExpression:
            return new FunctionExpressionHandler();
        case ts_morph_1.SyntaxKind.ArrowFunction:
            return new ArrowFunctionHandler();
        case ts_morph_1.SyntaxKind.FunctionDeclaration:
            return new FunctionDeclarationHandler();
        case ts_morph_1.SyntaxKind.Constructor:
            return new ConstructorDeclarationHandler();
        case ts_morph_1.SyntaxKind.MethodDeclaration:
            return new MethodDeclarationHandler();
        case ts_morph_1.SyntaxKind.MethodSignature:
            return new MethodSignatureHandler();
        case ts_morph_1.SyntaxKind.ClassDeclaration:
            return new ClassDeclarationHandler();
        case ts_morph_1.SyntaxKind.InterfaceDeclaration:
            return new InterfaceDeclarationHandler();
        case ts_morph_1.SyntaxKind.EnumDeclaration:
            return new EnumDeclarationHandler();
        case ts_morph_1.SyntaxKind.PropertyDeclaration:
            return new PropertyDeclarationHandler();
        case ts_morph_1.SyntaxKind.PropertySignature:
            return new PropertySignatureHandler();
        case ts_morph_1.SyntaxKind.VariableDeclaration:
            return new VariableDeclarationHandler();
        case ts_morph_1.SyntaxKind.GetAccessor:
            return new GetAccessorHandler();
        case ts_morph_1.SyntaxKind.SetAccessor:
            return new SetAccessorHandler();
        case ts_morph_1.SyntaxKind.EnumMember:
            return new EnumMemberHandler();
        case ts_morph_1.SyntaxKind.ModuleDeclaration:
            return new NamespaceHandler();
        case ts_morph_1.SyntaxKind.TypeAliasDeclaration:
            return new TypeAliasHandler();
        case ts_morph_1.SyntaxKind.PropertyAssignment:
            return new PropertyAssignmentHandler();
        default:
            return null;
    }
}
function traverseNodes(node) {
    var handler = createHandler(node);
    if (handler) {
        handler.enter(node);
        node.forEachChild(function (n) { return traverseNodes(n); });
        parentStack.pop();
    }
    else if (node.getKind() != ts_morph_1.SyntaxKind.ExportDeclaration && node.getKind() != ts_morph_1.SyntaxKind.ImportDeclaration && node.getKind() != ts_morph_1.SyntaxKind.ArrowFunction) {
        node.forEachChild(function (n) { return traverseNodes(n); });
    }
}
function isAssignmentToken(token) {
    switch (token.getKind()) {
        case ts_morph_1.SyntaxKind.EqualsToken:
        case ts_morph_1.SyntaxKind.PlusEqualsToken:
        case ts_morph_1.SyntaxKind.MinusEqualsToken:
        case ts_morph_1.SyntaxKind.AsteriskAsteriskEqualsToken:
        case ts_morph_1.SyntaxKind.AsteriskEqualsToken:
        case ts_morph_1.SyntaxKind.SlashEqualsToken:
        case ts_morph_1.SyntaxKind.PercentEqualsToken:
        case ts_morph_1.SyntaxKind.AmpersandEqualsToken:
        case ts_morph_1.SyntaxKind.BarEqualsToken:
        case ts_morph_1.SyntaxKind.CaretEqualsToken:
        case ts_morph_1.SyntaxKind.LessThanLessThanEqualsToken:
        case ts_morph_1.SyntaxKind.GreaterThanGreaterThanGreaterThanEqualsToken:
        case ts_morph_1.SyntaxKind.GreaterThanGreaterThanEqualsToken:
        case ts_morph_1.SyntaxKind.BarBarEqualsToken:
        case ts_morph_1.SyntaxKind.AmpersandAmpersandEqualsToken:
        case ts_morph_1.SyntaxKind.QuestionQuestionEqualsToken:
            return true;
    }
    return false;
}
function addDependency(from, to, kind, line) {
    root.addDependency(from, to, line, kind);
}
var fileMap = new Map();
function processFunctionOrMethod(decl, parent, sym, parameters, isAbstract, type) {
    return findOrCreateModelElement(parent, sym.getName(), decl, type, function (parent, name, type, decl, line) { return new model_1.FunctionOrMethodWithoutBody(parent, name, decl, line, type, isAbstract); }, function (element) { return handleParameterTypes(element, parameters, false); });
}
function findOrCreateModelElement(parent, name, decl, type, factory, afterCreationCallback) {
    var line = decl.getStartLineNumber();
    var found = (0, model_1.lookupNode)(decl);
    if (!found) {
        found = factory(parent, name, type, decl, line);
        if (afterCreationCallback) {
            afterCreationCallback(found);
        }
    }
    return found;
}
function processDeclaration(decl, sourceFileElement) {
    var sym = decl.getSymbol();
    if (!sym) {
        if (decl.getParent())
            return processDeclaration(decl.getParent(), sourceFileElement);
        return sourceFileElement;
    }
    var flags = sym.getFlags();
    var parent;
    if (decl.getParent())
        parent = processDeclaration(decl.getParent(), sourceFileElement);
    else
        parent = sourceFileElement;
    if (!parent)
        return undefined;
    if (decl instanceof ts_morph_1.SourceFile)
        return sourceFileElement;
    if (flags & ts_morph_1.SymbolFlags.Namespace)
        return findOrCreateModelElement(parent, sym.getName(), decl, model_1.ElementType.Namespace, function (parent, name, type, decl, line) { return new model_1.Namespace(parent, name, decl, line); });
    if (flags & ts_morph_1.SymbolFlags.Class) {
        var isAbstract_1 = false;
        var name_1 = sym.getName();
        if (decl instanceof ts_morph_1.ClassDeclaration) {
            var cd = decl;
            isAbstract_1 = cd.isAbstract();
            if (name_1 === "default" && cd.getName())
                name_1 = cd.getName();
        }
        return findOrCreateModelElement(parent, name_1, decl, model_1.ElementType.Class, function (parent, name, type, decl, line) { return new model_1.Class(parent, name, decl, line, isAbstract_1); });
    }
    if (flags & ts_morph_1.SymbolFlags.Interface)
        return findOrCreateModelElement(parent, sym.getName(), decl, model_1.ElementType.Interface, function (parent, name, type, decl, line) { return new model_1.Interface(parent, name, decl, line); });
    if (flags & ts_morph_1.SymbolFlags.Function) {
        if (decl instanceof ts_morph_1.FunctionDeclaration) {
            var functionNode = decl;
            return processFunctionOrMethod(decl, parent, sym, functionNode.getParameters(), false, model_1.ElementType.Function);
        }
        return undefined;
    }
    if (flags & ts_morph_1.SymbolFlags.Property)
        return findOrCreateModelElement(parent, sym.getName(), decl, model_1.ElementType.Property, function (parent, name, type, decl, line) { return new model_1.PropertyOrVariable(parent, name, decl, line, type); });
    if (flags & ts_morph_1.SymbolFlags.Method) {
        if (decl instanceof ts_morph_1.MethodDeclaration) {
            var methodNode = decl;
            return processFunctionOrMethod(decl, parent, sym, methodNode.getParameters(), methodNode.isAbstract(), model_1.ElementType.Method);
        }
        if (decl instanceof ts_morph_1.MethodSignature) {
            var methodNode = decl;
            return processFunctionOrMethod(decl, parent, sym, methodNode.getParameters(), false, model_1.ElementType.Method);
        }
        return undefined;
    }
    if (flags & ts_morph_1.SymbolFlags.Variable)
        return findOrCreateModelElement(parent, sym.getName(), decl, model_1.ElementType.Variable, function (parent, name, type, decl, line) { return new model_1.PropertyOrVariable(parent, name, decl, line, type); });
    if (flags & ts_morph_1.SymbolFlags.Enum)
        return findOrCreateModelElement(parent, sym.getName(), decl, model_1.ElementType.Enum, function (parent, name, type, decl, line) { return new model_1.Enum(parent, name, decl, line); });
    if (flags & ts_morph_1.SymbolFlags.EnumMember)
        return findOrCreateModelElement(parent, sym.getName(), decl, model_1.ElementType.EnumMember, function (parent, name, type, decl, line) { return new model_1.PropertyOrVariable(parent, name, decl, line, type); });
    if (flags & ts_morph_1.SymbolFlags.GetAccessor) {
        var getterNode = decl;
        return processFunctionOrMethod(decl, parent, sym, getterNode.getParameters(), false, model_1.ElementType.Getter);
    }
    if (flags & ts_morph_1.SymbolFlags.SetAccessor) {
        var setterNode = decl;
        return processFunctionOrMethod(decl, parent, sym, setterNode.getParameters(), false, model_1.ElementType.Setter);
    }
    if (flags & ts_morph_1.SymbolFlags.Constructor) {
        var constructorDecl = decl;
        return processFunctionOrMethod(decl, parent, sym, constructorDecl.getParameters(), false, model_1.ElementType.Setter);
    }
    if (flags & ts_morph_1.SymbolFlags.TypeAlias)
        return findOrCreateModelElement(parent, sym.getName(), decl, model_1.ElementType.TypeAlias, function (parent, name, type, decl, line) { return new model_1.TypeAlias(parent, name, decl, line); });
    return undefined;
}
function addDependencyToSymbol(from, sym, line, kind) {
    // const originalSym = sym;
    while (sym.getAliasedSymbol())
        sym = sym.getAliasedSymbol();
    if (sym.getFlags() == ts_morph_1.SymbolFlags.ValueModule)
        return;
    var to = (0, model_1.lookupSymbol)(sym);
    if (to) {
        if (to instanceof model_1.Enum && kind == model_1.DependencyKind.READ) {
            return;
        }
        addDependency(from, to, kind, line);
    }
    else if (!localSymbols.has(sym) && (sym.getFlags() & (ts_morph_1.SymbolFlags.Prototype | ts_morph_1.SymbolFlags.BlockScopedVariable | ts_morph_1.SymbolFlags.FunctionScopedVariable)) == 0) {
        // External symbol
        var fqn = sym.getFullyQualifiedName();
        if (!fqn.startsWith("__") && sym.getDeclarations().length > 0) {
            var decl = sym.getDeclarations()[0];
            to = (0, model_1.lookupNode)(decl);
            if (!to) {
                var path = decl.getSourceFile().getFilePath();
                if (!path)
                    return;
                var sourceFileElement = fileMap.get(path);
                if (!sourceFileElement) {
                    sourceFileElement = new model_1.SourceFileElement(root, path, decl.getSourceFile(), true);
                    fileMap.set(path, sourceFileElement);
                }
                to = processDeclaration(decl, sourceFileElement);
            }
            if (to) {
                addDependency(from, to, kind, line);
            }
        }
    }
}
function addDependencyToNode(from, toNode, kind) {
    var sym = toNode.getSymbol();
    if (sym) {
        addDependencyToSymbol(from, sym, toNode.getStartLineNumber(), kind);
    }
}
function addDependencyToType(from, type, visitedTypes, line, dependencyKind) {
    if (dependencyKind === void 0) { dependencyKind = model_1.DependencyKind.USES; }
    if (visitedTypes.has(type))
        return;
    visitedTypes.add(type);
    if (type.isUnion()) {
        type.getUnionTypes().forEach(function (ut) { return addDependencyToType(from, ut, visitedTypes, line, dependencyKind); });
    }
    else if (type.isIntersection()) {
        type.getIntersectionTypes().forEach(function (it) { return addDependencyToType(from, it, visitedTypes, line, dependencyKind); });
    }
    else if (type.isTuple()) {
        type.getTupleElements().forEach(function (te) { return addDependencyToType(from, te, visitedTypes, line, dependencyKind); });
    }
    else if (type.isArray()) {
        addDependencyToType(from, type.getArrayElementType(), visitedTypes, line, dependencyKind);
    }
    else if (type.isClassOrInterface()) {
        var sym = type.getSymbol();
        if (sym) {
            addDependencyToSymbol(from, sym, line, dependencyKind);
        }
        type.getTypeArguments().forEach(function (ta) { return addDependencyToType(from, ta, visitedTypes, line, dependencyKind); });
    }
}
function isIdentifier(node) {
    return node.getKind() == ts_morph_1.SyntaxKind.Identifier || node.getKind() == ts_morph_1.SyntaxKind.PrivateIdentifier;
}
function handleBinaryExpression(node, forbiddenNodes, from) {
    var isAssignmentOperator = isAssignmentToken(node.getOperatorToken());
    var leftNode = node.getChildAtIndex(0);
    var rightNode = node.getChildAtIndex(2);
    if (isAssignmentOperator && isIdentifier(leftNode)) {
        addDependencyToNode(from, leftNode, model_1.DependencyKind.WRITE);
    }
    else {
        detectDependencies(leftNode, forbiddenNodes, from, isAssignmentOperator ? model_1.DependencyKind.WRITE : model_1.DependencyKind.READ);
    }
    detectDependencies(rightNode, forbiddenNodes, from, model_1.DependencyKind.READ);
}
function handlePropertyAccess(node, forbiddenNodes, from, dependencyKind) {
    var leftNode = node.getChildAtIndex(0);
    var rightNode = node.getChildAtIndex(2);
    if (isIdentifier(leftNode)) {
        addDependencyToNode(from, leftNode, model_1.DependencyKind.READ);
    }
    else {
        detectDependencies(leftNode, forbiddenNodes, from, model_1.DependencyKind.READ);
    }
    if (isIdentifier(rightNode)) {
        addDependencyToNode(from, rightNode, dependencyKind);
    }
}
function handleCallExpression(node, forbiddenNodes, from) {
    var expr = node.getExpression();
    if (isIdentifier(expr)) {
        addDependencyToNode(from, expr, model_1.DependencyKind.CALL);
    }
    else {
        detectDependencies(node.getExpression(), forbiddenNodes, from, model_1.DependencyKind.CALL);
    }
    node.getArguments().forEach(function (arg) { return detectDependencies(arg, forbiddenNodes, from, model_1.DependencyKind.READ); });
}
function handlePostfixUnaryExpression(node, forbiddenNodes, from, dependencyKind) {
    var operand = node.getOperand();
    if (isIdentifier(operand)) {
        addDependencyToNode(from, operand, model_1.DependencyKind.WRITE);
    }
    else {
        detectDependencies(operand, forbiddenNodes, from, dependencyKind);
    }
}
function handlePrefixUnaryExpression(node, forbiddenNodes, from) {
    var operator = node.getOperatorToken();
    var operand = node.getOperand();
    var isIncOrDec = operator == ts_morph_1.SyntaxKind.PlusPlusToken || operator == ts_morph_1.SyntaxKind.MinusMinusToken;
    if (isIncOrDec && isIdentifier(operand)) {
        addDependencyToNode(from, operand, model_1.DependencyKind.WRITE);
    }
    else {
        detectDependencies(operand, forbiddenNodes, from, isIncOrDec ? model_1.DependencyKind.WRITE : model_1.DependencyKind.READ);
    }
}
function handleNewExpression(node, forbiddenNodes, from) {
    var expr = node.getExpression();
    if (isIdentifier(expr)) {
        addDependencyToNode(from, expr, model_1.DependencyKind.NEW);
    }
    else {
        // Likely PropertyAccessExpression
        detectDependencies(expr, forbiddenNodes, from, model_1.DependencyKind.NEW);
    }
    node.getArguments().forEach(function (arg) { return detectDependencies(arg, forbiddenNodes, from, model_1.DependencyKind.READ); });
}
function handleTypeReference(node, forbiddenNodes, from) {
    addDependencyToType(from, node.getTypeName().getType(), new Set(), node.getStartLineNumber());
}
function detectDependencies(node, forbiddenNodes, from, dependencyKind) {
    if (forbiddenNodes.has(node))
        return;
    switch (node.getKind()) {
        case ts_morph_1.SyntaxKind.BinaryExpression:
            handleBinaryExpression(node, forbiddenNodes, from);
            break;
        case ts_morph_1.SyntaxKind.PropertyAccessExpression:
            handlePropertyAccess(node, forbiddenNodes, from, dependencyKind);
            break;
        case ts_morph_1.SyntaxKind.CallExpression:
            handleCallExpression(node, forbiddenNodes, from);
            break;
        case ts_morph_1.SyntaxKind.PostfixUnaryExpression:
            handlePostfixUnaryExpression(node, forbiddenNodes, from, model_1.DependencyKind.READ);
            break;
        case ts_morph_1.SyntaxKind.PrefixUnaryExpression:
            handlePrefixUnaryExpression(node, forbiddenNodes, from);
            break;
        case ts_morph_1.SyntaxKind.NewExpression:
            handleNewExpression(node, forbiddenNodes, from);
            break;
        case ts_morph_1.SyntaxKind.Identifier:
        case ts_morph_1.SyntaxKind.PrivateIdentifier:
            addDependencyToNode(from, node, model_1.DependencyKind.READ);
            break;
        case ts_morph_1.SyntaxKind.TypeReference:
            handleTypeReference(node, forbiddenNodes, from);
            break;
        case ts_morph_1.SyntaxKind.ExportDeclaration:
        case ts_morph_1.SyntaxKind.ImportDeclaration:
            break;
        case ts_morph_1.SyntaxKind.ModuleDeclaration:
            var md = node;
            if (md.getBody())
                detectDependencies(md.getBody(), forbiddenNodes, from, dependencyKind);
            break;
        case ts_morph_1.SyntaxKind.TypeParameter:
            var decl = node;
            if (decl.getConstraint()) {
                detectDependencies(decl.getConstraint(), forbiddenNodes, from, dependencyKind);
            }
            break;
        case ts_morph_1.SyntaxKind.PropertyAssignment:
            var pa = node;
            if (pa.getInitializer())
                detectDependencies(pa.getInitializer(), forbiddenNodes, from, dependencyKind);
            break;
        case ts_morph_1.SyntaxKind.FunctionDeclaration:
            var fd = node;
            if (fd.getBody())
                detectDependencies(fd.getBody(), forbiddenNodes, from, dependencyKind);
            break;
        case ts_morph_1.SyntaxKind.ArrowFunction:
            var af = node;
            if (af.getBody())
                detectDependencies(af.getBody(), forbiddenNodes, from, dependencyKind);
            break;
        default:
            node.forEachChild(function (cn) { return detectDependencies(cn, forbiddenNodes, from, dependencyKind); });
            break;
    }
}
function visitSourceFiles(languageService, tsConfigPath, sourceFiles) {
    root = new model_1.Root(tsConfigPath);
    var _loop_1 = function (sourceFile) {
        var path = sourceFile.getFilePath();
        var sf = new model_1.SourceFileElement(root, path, sourceFile);
        var hasStatements = false;
        fileMap.set(path, sf);
        parentStack.push(sf);
        sourceFile.forEachChild(function (n) {
            traverseNodes(n);
            if (n.getKind() >= ts_morph_1.SyntaxKind.FirstStatement && n.getKind() <= ts_morph_1.SyntaxKind.LastStatement && n.getKind() != ts_morph_1.SyntaxKind.VariableStatement)
                hasStatements = true;
        });
        if (hasStatements) {
            var mc = new metric_collector_1.MetricCollector(sourceFile);
            mc.collectMetrics();
            sf.numberOfStatements = mc.numberOfStatements;
            sf.cyclomaticComplexity = mc.cyclomaticComplexity;
            sf.modifiedCyclomaticComplexity = mc.modifiedCyclomaticComplexity;
            sf.logicalOperations = mc.logicalOperations;
            sf.maxIndent = mc.maxIndent;
        }
        parentStack.pop();
    };
    for (var _i = 0, sourceFiles_1 = sourceFiles; _i < sourceFiles_1.length; _i++) {
        var sourceFile = sourceFiles_1[_i];
        _loop_1(sourceFile);
    }
    delayedActions.forEach(function (action) { return action(); });
    delayedActions = [];
    var elements = (0, model_1.getAllElements)();
    var upperIndex = elements.length;
    var _loop_2 = function (i) {
        var element = elements[i];
        if (element instanceof model_1.FunctionOrMethodWithBody) {
            var childrenNodes = new Set();
            for (var _a = 0, _b = element.children; _a < _b.length; _a++) {
                var child = _b[_a];
                if (!(child instanceof model_1.PropertyOrVariable)) {
                    childrenNodes.add(child.node);
                }
            }
            if (element.bodyNode) {
                detectDependencies(element.bodyNode, childrenNodes, element, model_1.DependencyKind.READ);
            }
        }
        else if (element instanceof model_1.PropertyOrVariable) {
            if (element.initializer)
                detectDependencies(element.initializer, emptySet, element.parent, model_1.DependencyKind.READ);
        }
        else if (element instanceof model_1.SourceFileElement) {
            var childrenNodes_1 = new Set();
            element.children.forEach(function (c) { return childrenNodes_1.add(c.node); });
            detectDependencies(element.node, childrenNodes_1, element, model_1.DependencyKind.READ);
        }
    };
    for (var i = 0; i < upperIndex; i++) {
        _loop_2(i);
    }
    return root;
}
exports.visitSourceFiles = visitSourceFiles;
//# sourceMappingURL=visitor.js.map